(* NOF a Small File System for Flash memory using FID and DID hash tables

  FID = File ID length = 8 bytes

    | sector number | dir hash | filename hash | file length |  ( max. 384 files )


  The hash-code may be any value from 1 to FE. Zero and FF are
  invalid hash-codes. Hash code 0 means an erase entry! Hash
  code FF means an unused DID-entry.

  DID = Directory ID length = 16 bytes ( max. 64 directories )

    | Dir-hash | Working-Dir-hash | Dir-name string |


  File storage, first sector:

      Byte Byte cell       28 bytes
    ---------------------------------------------------
    | FIT | Aux | XT | Filename 28 bytes | File data 


  FIle Types (FIT):

    0 = TI          - TI TXT file
    1 = INTELHEX    - Intel HEX file
    2 = ASCII       - Plain text file
    3 = FORTH       - Forth source file
    4 - Pictures    - Picture bitmap file

*)

v: forth definitions  inside also
: D-        ( dn1 dn2 -- dn3 )          dnegate d+ ;
\ : UM+       ( ud1 u1 -- ud2 )           0 d+ ;

: RECURSE       ( -- )      \ Recursive noForth call
    ?comp s0 @ dup if  , exit  then  1- ?abort ; immediate


\ Write to sector buffers, administrate & write a sector to Flash
\ Store b in active sector buffer
\ 0 = Write flag
\ 1 = Length of binary string
\ 2 = Sector number to patch
\ 3 = Position in sector to patch
\ 4 = FID sector to patch for file length
\ 5 = FID record address
v: inside definitions
create RWDATA  6 cells allot \ R/W data info
\ : !DATA     ( x +n -- )     cells rwdata + ! ;
\ : @DATA     ( +n -- x )     cells rwdata + @ ;
\ : B,    ( b -- )
\    #n 100 = if  true 0 !data  1 to buf  0 to #n  then
\    buf buffer #n + c!  incr #n ;  \ Store byte

code !DATA  ( x +n -- )
    5707 ,                  \ tos  tos add
    5037 , rwdata ,         \ rwdata #  tos add
    44B7 , 0 ,              \ sp )+  tos ) mov
    4437 , next             \ sp )+  tos mov
end-code

code @DATA  ( +n -- x )
    5707 ,                  \ tos  tos add
    5037 , rwdata ,         \ rwdata #  tos add
    4727 , next             \ tos )  tos mov
end-code

code B, ( b -- )
    4218 ,  adr #n ,        \ adr #n &  day mov       \ DAY = #n
    9038 , 100 , 2005 ,     \ 100 #  day cmp  =? if,  \ #n = 100 ?
    43B2 , rwdata ,         \     #-1  rwdata & mov   \ Yes, true flag to data+0
    4392 , adr buf ,        \     #1  adr buf & mov   \ Change to buffer-1
    F378 ,                  \     #-1 day .b bia      \ Restore #n to zero
                            \ then,
    421A , adr buf ,        \ adr buf &  sun mov      \ Get buf
    F37A , 108A ,           \ #-1 sun .b bia sun swpb \ 100 *
    503A , 'sector ,        \ 'sector #  sun add      \ Sector addr to sun
    580A ,                  \ day sun add             \ Add byte position #n
    47CA , 0 ,              \ tos  sun ) .b mov       \ Store byte in buffer
    5318 ,                  \ #1  day add             \ Incr #n
    4882 , adr #n ,         \ day  adr #n & mov       \ Save new #n
    4437 , next             \ sp )+  tos mov   next   \ drop 'b' and ready
end-code

: WR-BUFFER     ( -- )      \ Write buffer-0
    ." #"  #sector  0 write-sector  incr #sector ; \ Save buffer 0

: <WRITE>       ( -- )      \ Write buffer 0 when it's overflowed
    0 @data if
        wr-buffer  false 0 !data    \ Write done
        1 buffer  0 buffer  #n move  0 to buf   \ Restore receive buffer
    then ;

: INIT-RWDATA   ( -- )  \ Correct start for buffer system
    0 0 !data  0 1 !data \ Reset RW-record
    0 to #n  0 to buf ; \ Use buffer 0 from start

: !LENGTH       ( -- )  \ Patch file length in a 256 byte FID sectors
    4 @data 0 read-sector       \ Read FID sector to patch
    5 @data 0 buffer + >r       \ Get FID entry address
    #sector  r@ @ -  dup u.  r> 6 + ! \ Calc. length & store in last position of FID
    4 @data 0 write-sector ;    \ Write FID sector back


0000 constant #FID      \ NOF block with File-ID block (384 files)
0C00 constant #DID      \ and Directory-ID block (64 dirs)

create 'NAME  20 allot  \ Hold name string to convert
10 to #sector           \ First free addressable sector
value DID   0A allot    \ Hold DIR hash code & nesting
value 'DID              \ Free dir pointer
value DPT               \ DID nesting pointer
\ value FID               \ Hold FILE hash code
value 'FID              \ Free file pointer
value #HASH             \ Temporary HASH memory
value TMP               \ Temporary scratch memory


: ROOT      ( -- )      0 to DID  0 to DPT ;    \ Set root dir

: >DID      ( did -- )  \ Add new top DIR to path
    DPT 0A = ?abort  2 +to DPT
    adr did dup cell+ DPT move  to DID ;

: DID>      ( -- )      \ Remove top DIR from path
    DPT if  adr did dup cell+ swap  DPT move  -2 +to DPT  then ;

: MDID      ( -- h )    \ Generate a 16-bit DIR hash code
    #hash DID FF and b+b ( dup u. ) ;

: $>NAME    ( a0 u -- a1 u ) \ Make all file strings uppercase
    >r  r@ 'name c!  'name 1+ r> move \ Copy name to 'NAME
    'name count  2dup upper ;       \ Convert to uppercase


\ A KEY to read from Flash disk

create 'FADDR  4 allot      \ Double Flash byte-address
code FADDR  ( -- fa )
    8324 , 4784 , 0 ,       \ tos sp -) mov           \ Save TOS
    4217 , 'faddr ,         \ 'faddr & tos mov        \ High cell to TOS
    8324 , 4294 , 'faddr 2 + , 0 , \ 'faddr 2 + & sp -) mov  \ Push low cell to stack
next  end-code
code !FADDR ( d -- )
    4782 ,  'faddr ,        \ tos 'faddr & mov        \ Store Hi
    44B2 ,  'faddr 2 + ,    \ sp )+ 'faddr 2 + & mov  \ Store Lo
    4437 ,                  \ sp )+ tos mov           \ Pop stack
next  end-code
code INC-FADDR ( -- )
    5392 , 'faddr 2 + ,    \ #1 'faddr 2 + & add     \ Incr. low
    2002 , 5392 , 'faddr , \ =? if,  #1 'faddr & add  then, \ Overflow? incr. Hi
next  end-code

: FKEY      ( -- c )        \ File KEY
    faddr fc@ inc-faddr ?dup ?exit -2 to source-id 0D ;


v: extra definitions
: MOUNT     ( -- )          \ Scan NOF to intialise disk
    spi-setup  init-rwdata
    ['] fkey  adr 'key cell+ ! \ Replace second KEY
\ Initialise FID and DID to next free location
    #FID  begin  dup 0 f@ -1 <> while  dup to TMP  8 +   repeat  to 'FID
    #DID  begin  dup 0 f@ -1 <> while  10 +  repeat  to 'DID
\ Initialise first free sector location to #SECTOR
    #FID 'FID <> if         \ FID used?
        TMP 0 f@  TMP 6 + 0 f@ + \ Yes, calc. next free sector
    else  10  then          \ No, first free sector
    to #sector  ROOT ;      \ Init. it, start in ROOT



\ Build a FID table in sector-0 using DHASH
v: inside definitions
code DHASH      ( a n -- h )    \ h = byte
    4708 , 443A ,   \ tos day mov   sp )+ sun mov
                    \ begin, 
    5747 ,          \    tos tos .b add
    6347 ,          \    #0 tos .b addc
    EA77 ,          \    sun )+ tos .b bix
    8318 ,          \    #1 day sub
    23FB , next     \ 0=? until,  next
end-code

\ Check hash h in current DIR, leave true when hash value found, otherwise false
: CHECK-DHASH   ( h -- fl )         \ h = Byte wide hash
    'DID to TMP                     \ Save current dir-entry address
    #DID 'DID over -  bounds ?do    \ Check directory space
        dup i 0 f@ = if             \ and for hash match, leave loop if so
            i to TMP                \ Save dir address 
            drop  true  unloop exit \ Drop hash & leave true
        then
    10 +loop  drop  false ;         \ To next DIR entry, or no match

\ Display disk & folder information
: .DIR          ( did -- )  \ Show folder name
    ?dup 0= if  ." Root\"  exit  then
    check-dhash if
        TMP cell+ 0 fc@+ ftype ." \"  \ Yes, show name
    then ;

: .PATH         ( -- )      \ Show current folder path
    adr DID  dup DPT +  do  i @ .dir  -2 +loop ; 

: >DHASH        ( a u -- )          \ Generate and save hash code from the string a u
    0D umin $>name  dhash to #hash ; \ Save & limit name, Calc. & save hash

s" .." dhash constant ..    \ Hash code for ..
s" \"  dhash constant \\    \ Hash code for \
: MAKE-DIR      ( a u -- )
    >dhash #hash .. = #hash \\ = or \ Check .. or \
    #hash 0= or  #hash FF = or      \ and zero of FF, finally
    MDID check-dhash  or ?abort     \ check the double-byte hash
    1000 'DID - 10 < ?abort         \ Still room for directory
    'DID 100 /mod 0 read-sector     \ Read active dir sector
    0 buffer + >r  MDID r@ !        \ Fill in dir hash codes
    'name dup c@ 1+  r> cell+ swap move \ Store name of new dir
    'DID 100 / 0 write-sector       \ Write dir sector back
    10 +to 'DID ;                   \ To next dir entry

v: extra definitions
: MD            ( "name" -- )           bl-word count  make-dir  cr .path ;


v: inside definitions
: CHOOSE-DIR    ( a u -- )  \ Generate hash & select directory
    >dhash  #hash \\ = if  ROOT exit  then \ Back to dmp root on \
    #hash .. = if  did> exit  then  \ DIR level back on ..
    MDID check-dhash 0= ?abort      \ Abort when DIR does not exists
    MDID >DID ;                     \ Extend DIR nesting

v: extra definitions
\ CD name   select DIR name  
\ CD ..     Go DIR level back
\ CD \      Go to root dir
: CD            ( "name" -- )           bl-word count  choose-dir  cr .path ;


\ Build a FID table in sector-0
v: inside definitions
code FHASH      ( a n -- hh )   \ hh = cell
    4708 , 443A ,       \ tos day mov   sp )+ sun mov
                        \ begin,
    5707 ,              \    tos tos add
    6307 , 1087 ,       \    #0 tos addc   tos swpb
    4A76 , E607 ,       \    sun )+ w .b mov   w tos bix
    8318 ,              \    #1 day sub   
    23F9 , next         \ 0=? until,  next
end-code

\ The value TMP contains the start sector of the file
\ Leave true when the hash-value hh is found, otherwise false
: CHECK-FHASH   ( hh -- fl )
    #FID 'FID bounds ?do            \ Scan used FID
        i 2 + 0 f@ DID = if         \ Current DIR?
            dup i 4 + 0 f@ = if     \ Yes, wanted file
                i to TMP            \ Remember FID record
                drop true unloop  exit \ & ready
            then
        then
    8 +loop                         \ Next entry, or done
    drop  #FID to TMP  false ;      \ No FID record found

: FILE-NAME     ( xt #fit a u -- )
    init-rwdata  1B umin $>name     \ Save & limit name
    fhash dup check-fhash ?abort    \ Calc. & check hash value
    #DID 'FID - 8 < ?abort          \ Still room for file entry?
    'FID 100 /mod  dup 4 !data      \ Save FID sector
    0 read-sector                   \ Read active FID sector
    dup 5 !data  0 buffer +         \ Get & save FID record address
    #sector  over !                 \ Save file location
    DID over 2 + !   4 + !          \ Save directory-name hash & file-name hash
    4 @data 0 write-sector          \ Write FID sector back
    8 +to 'FID                      \ & to next entry   
    #sector 0 read-sector           \ Read first sector of file
    0 buffer >r  FF00 or r@ !  r@ cell+ ! \ Store file-type & token
    'name dup c@ 1+  r> 2 cells +   \ Get filename
    swap move  20 to #n ;           \ Store name & set data start position!

v: extra definitions
\ #sector - DID - FID -
: ADD-FID       ( xt #fit "name" -- )   bl-word count  file-name ; 


v: inside definitions
: .FOLDERS      ( -- )      \ Show folders in current folder
    #DID 'DID over - bounds ?do
        i 1+ 0 fc@  did ff and = if
            cr  ." <dir> "  i cell+ 0 fc@+ ftype
        then
    10 +loop ;

: .FIT          ( +n -- )           \ Show FIle Type
    4 umin 3 * s" IntIhxAscFrtPic" drop + 3 type 2 spaces ;

v: extra definitions
: DIR           ( -- )      \ Show all current folder info
    cr ." Dir shown = " .path
    .folders
    #FID 'FID bounds ?do            \ Scan used FID
        i 2 + 0 f@  DID = if        \ Current DIR?
            i 0 f@ ?dup if
                cr  dup 4 u.r space \ Show sector
                100 um*  2dup fc@ .fit \ Show file type
                2 cells m+ fc@+ ftype \ & file name
            then
        then
    8 +loop  cr ." Free " .dfree ;  \ Next entry, or done

: ERASE-DISK ( -- )             \ Empty Flash completely!
    chip-erase  mount ;         \ Restore pointers

v: fresh definitions
' mount  to app
shield NOF\  freeze



(* Added file functionality

    MOUNT       - Add Flash drive to noForth ... Ok
    DROP-FILE   - Remove last added file ....... Ok
    DROP-FILES  - Remove all files ............. Ok
    ERASE-DISK  - Erase & init. a whole disk ... Ok
    DROP-DIR    - Remove last added folder ..... To-do
    TREE        - Show directory tree .......... Ok
    PROGRAM     - Program an MSP430 by name .... Ok
    INCLUDE     - Include a Forth file by name . Ok
    VIEW        - Show any file type bye name .. Ok
    DIR         - Show current directory ....... Ok
    MD          - Make directory name .......... Ok
    CD          - Choose directory name ........ Ok
    ASCII:      - Add a named ASCII file ....... Ok
    FORTH:      - Add a named Forth file ....... Ok
    TI:         - Add a named Intel TXT file ... Ok
    INTELHEX:   - Add named Intel-Hex file ..... Ok
    LOADFILE    - Load file from sector ........ Ok
    >VIEW       - Show any file type from sector Ok
    END-OF-FILE - Stop loading of a file (dummy) Ok

*)

v: inside also  definitions
: DROP-FILE)    ( -- )      \ Remove last saved file from Flash disk
    'FID 8 - >r             \ Last FID entry
    r@ 0 f@  to #sector     \ Release removed sectors
    #sector r@ 6 + 0 f@ bounds ?do
        i u. i erase-sector \ Erase 10 sectors each time
    10 +loop  r> to 'FID ;  

v: extra definitions
: DROP-FILE     ( -- )      \ Remove file, make FID invalid
    'FID #FID = ?exit       \ Do nothing when FID = empty
    drop-file)  'FID        \ Release file sectors, get last FID entry
    dup 100 / 0 read-sector  \ Read active FID sector
    dup 100 mod 0 buffer + 8 0 fill \ Set FID as erased
    100 / 0 write-sector  mount ; \ Write modified FID back

: DROP-FILES   ( -- )       \ Remove all files from Flash disk
    'FID #FID <> if         \ Do nothing when FID = empty
        'FID 8 / for  drop-file)  next
    then
    0 erase-sector  mount ;


\ Interpret files fom Flash disk, zero is EOF!
v: inside definitions
: >FHASH        ( a u -- )          \ Generate and save hash code from the string a u
    1B umin $>name  fhash to #hash ; \ Save & limit name, Calc. & save hash

: 'SEEK?    ( a u -- sector true | false ) \ Search NOF for string a u
    >fhash                  \ Convert name to 16-bit hash
    #hash check-fhash if    \ Check if hash exists
        TMP 0 f@  true  exit
    then  ."  File not found "  false ;

: SEEK?     ( "name" -- sector true | false )   \ Search NOF for "name"
    bl-word count 'seek? ;

v: extra definitions
: END-OF-FILE ( -- ) ;          \ Stop loading a file (Dummy).

: LOADFILE  ( sector -- )       \ Load file from sector address
    100 um* 20 M+   source-id   \ Calc. file start addr, get s-id
    dup 0 > if  faddr 2>r  then \ Are we loading a file?
    >r  !faddr  1 to source-id  \ Save s-id too, set new fileaddr.
    interpret  r@ to source-id  \ Interpret & restore s-id
    r> 0 > if  2r> !faddr  then ; \ Restore faddr, when used.

v: forth definitions
: INCLUDE   ( "name" -- )       \ Include Forth file
    state @ ?abort  seek? if    \ Interactive only
        >r  r@ 100 um* fc@      \ Get file startsector & FIT  
        3 <> ?abort  r>         \ Check FIT? get sector number
        loadfile
    then ;  immediate


\ Show ASCII & binary data as hexdump
v: inside definitions
: SHOW)     ( sector -- )   \ Type ASCI file from Flash
    100 um* 20 m+  0 >r cr  \ Save line counter
    begin
        fc@+ dup 0D = if    \ EOL?
            r> 1+ dup 10 = if   \ Get & incr. line counter is it 16?
                key bl <> if    \ Quit on all keys but the space bar
                    2drop 2drop exit
                then  dup - \ Reset counter
            then  >r        \ Save counter
        then
    ?dup while
        emit
    repeat  r> drop  2drop ;

: PICDUMP   ( #sector -- )  \ Dump a picture file
    100 um*  2dup 1 m+ fc@  \ Number of pics
    >r  20 m+  r> 0 ?do     \ To start of data
        cr i .              \ Show picture number
        fc@+ >r fc@+ r> * >r \ Calc. picture size
        2dup r@ fdump  cr r> m+ \ Dump one pict. & to next
        key bl <> if  leave  then \ Stop when other then space
    loop
    2drop ;

: HEXDUMP   ( sector -- )   \ Display binary files
    100 um* 20 m+  begin
        2dup f@     \ da a
    while           \ da
        2dup  2dup  2 m+ f@  4 + fdump \ Dump binary
        key bl <> if  2drop exit  then  cr
        2 m+  2dup f@ m+ 2 m+ \ Skip binary string
    repeat  2drop ;

v: extra definitions
: >VIEW     ( sector -- )
    dup 100 um* fc@ >r      \ Read file type
    r@ 2 =  r@ 3 = or if    \ Text type?    
        rdrop  show)  exit  \ Yes, show that file
    then
    r> 4 = if  picdump  exit  then \ Picture file?
    hexdump ;               \ No, make hex dump

v: forth definitions
: VIEW      ( "name" -- )   \ Display any file
    seek? if  >view  then ;


v: inside definitions
: PROGRAM)  ( sector -- )
    100 um* 2>r  2r@ fc@ >r \ Fetch FIT from first sector
    r@ 0<>  r> 1 <> and ?abort \ Check FIT
    2r@ 20 m+               \ Calc. byte addr. of first byte
    2r> 2 m+ f@ execute ;   \ Execute token from first sector

\ Use HEX-file to do programming actions
v: forth definitions
: PROGRAM   ( "name" -- )       \ Use file using it's private token
    flyer  seek? 0= ?abort  postpone literal  postpone program) ;  immediate

v: fresh definitions
shield FILES\  freeze



(* New intel HEX read to Flash disk implementation
*)

v: inside also definitions
value #ADDR     \ Current address & checksum
value #CHK
: XKEY      ( -- c )    key  dup emit ;
: NUM>      ( -- n )    xkey upc 10 dig? 0= ?abort ; \ Convert char to number
: IHX>      ( -- b )    num> 10 *  num> or  dup +to #chk ; \ Hex byte to number
: CHK?      ( -- )      #chk negate ff and ihx> <> ?abort ; \ Checksum control
: ADDR>     ( -- a )    ihx> >< ihx> or ;       \ Hex word to address

: SKIPCOLON ( -- )      \ Wait for : char
    cr  20 0 do  xkey ch : = if unloop exit then  loop  true ?abort ;

: <PATCH>   ( -- )      \ Patch length in saved sector
    1 @data if                      \ Buffer used?
        ch * emit  2 @data .        \ Show patched sector
        2 @data 1 read-sector       \ Get sector in spare buffer
        1 @data  3 @data 1 buffer +  ! \ Patch length
        2 @data 1 write-sector  0 1 !data \ Save sector again
    then ;

: PATCH-SECTOR  ( #n #sector +n -- )
    1 !data  2 !data  3 !data   \ Save data
    #sector  2 @data = if       \ This sector?
        1 @data  3 @data  0 buffer +  ! \ Just, patch length
        0 1 !data               \ Done now, so clear length
    then ;

: NEWRECORD     ( da +n fl | 0 -1 -- da +n )
    if                          \ New record?  - 0 tf
        addr> dup #addr <> if   \ Address not continuously?  - 0 a
            ( New ) to #addr    \ Yes, note new record  - 0
            ?dup if patch-sector then \ Store prev. record length if any  -
            #addr b,  #addr >< b, \ Start new record  -
            #n #sector  FF b, FF b,  0  \ #n sa 0
        else
            drop
        then
    else
        patch-sector  0 b, 0 b, \ Store last record length & file end
    then ;

v: forth definitions
: INTELHEX:     ( xt "name" ccc -- )
    1 add-fid  -1 to #addr  0   \ Initialise address & record length
    begin
        skipcolon               \ Wait for new ihx-line - 0
        0 to #chk               \ Start new checksum - 0
        ihx> ?dup if            \ Get Intel HEX record length, not zero? - 0 +n
            >r  true newrecord  \ Finish record or start new! - 0 tf
            ihx> 0<> ?abort     \ Check if record type is valid
            r> 0 ?do            \ Read & store one line
                ihx> b,  1+  incr #addr \ Store data to record
            loop
            chk?  false         \ ."  CHK" \ Check checksum?
            <write>  <patch>    \ Save and/or patch sector
        else
            false newrecord     \ Finish last record
            ihx> ihx> 2drop     \ EOL, drop dummy address
            ihx> 1 =  chk?      \ Ready & check last checksum?
        then
    until  wr-buffer  <patch>
    adj-sector  !length  cr ."  End " ;


v: inside definitions
: EOL       ( c -- c )          \ At EOL do a write & patch if requested
    dup 0D = if  <write>  <patch>  then ;

: HKEY      ( -- b fl )     key  eol  10 dig? ; \ Handle EOL & try to convert a hex digit?
: BYTE      ( -- byte )     hkey 0= ?abort  10 *  hkey 0= ?abort  or ; \ Read TI-hex byte

v: forth definitions
: TI:       ( xt "name" ccc -- )
    0 add-fid   cr  false       \ Start new TI TXT file
    begin
        begin
            hkey 0=             \ Not a HEX-number?
        while
            dup ch q = if       \ Yes, ready when 'q' reached
                2drop  patch-sector  0 b,  0 b,
                wr-buffer  <patch>
                adj-sector  !length  ."  End"  exit
            else
                ch @ = if       \ No, start new code record?
                    if  patch-sector  then  \ Store record length
                    byte  byte b,  b,
                    #n #sector  \ Where to patch
                    FF b,  FF b,  0  true \ Empty record length
                then
            then
        repeat
        10 *  hkey 0= ?abort  or b,  >r  1+ \ Build number and store
        dup 10 mod 0= if  ch . emit  then  r> \ Show advance
    again ;


v: inside definitions
: BB,       ( c -- )        dup 0D = if  b,  0A  then  b, ;
: (ASCII)   ( "ccc" -- )            \ Store ASCII file in Flash
    cr  begin
        key  eol  dup  ch ; = if    \ First key, is it ';'
            key  eol  dup ch ; = if \ Second key, is it ';'
                key  eol  dup ch ; = if \ Third key, is it ';'
                    drop drop drop  space  \ Yes, file finished
                    0 b,  wr-buffer
                    adj-sector  !length  exit
                then
                rot b, swap b,      \ Third not ';' then store 3 chars
            else
                swap b,             \ Second not ';' store 2 chars
            then
        then  bb,                   \ Other char, store it
    again ;

v: forth definitions
: ASCII:   ( "name" ccc -- )      ['] >view     2 add-fid (ascii) ;
: FORTH:   ( "name" ccc -- )      ['] loadfile  3 add-fid (ascii) ;



(* Building files out of bitmap pictures

The first picture of a sequence of pictures starts with BITMAP:
Each next picture with PIC:
The last picture ends with BITMAP; this word end a picture
sequence end patches the file header with the number of
pictures stored and the length in #sectors.

    Commands:
        BITMAP:     ( xs ys "name " "ccc" -- )  Start bitmap movie sequence
        BITMAP;     ( -- )                      Finish bimap movie sequence
        PIC:        ( xs ys "ccc" -- )          Next bitmap picture

*)

v: inside also definitions
value #FIRST                    \ Remember first sector with header
value #PICS                     \ Remember number of pictures
: EOP       ( c -- c )          \ At EOP do a write if requested
    dup 0D = if  <write>  ch . emit  then ;

: PKEY      ( -- b fl )
   key  upc  eop  10 dig? ; \ Handle EOP & try to convert a hex digit?

v: extra definitions
: PIC:      ( xs ys ccc -- )    \ Next bitmap picture
    8 / dup b,  over b, *   \ Calc. byte count of bitmap
    cr  incr #pics  for     \ Count pictures
        begin  pkey 0= while drop repeat \ Leave a HEX-number?
        10 *  pkey 0= ?abort  or b, \ Build number and store
    next ;

: PICTURE;  ( -- )          \ Finish bitmap movie sequence 
    0 b,  0 b,  wr-buffer   \ Write picture end & last buffer  
    adj-sector  !length     \ Patch file length in FID
    #first  0 read-sector   \ Get File header
    #pics ><  0 buffer >r  r@ c@ or  r> ! \ Patch #pictures in file
    #first 0  write-sector  \ Replace file header
    cr ."  End" ;

: PICTURE:  ( xs ys "name" ccc -- )
    ['] 2drop  4 add-fid  0 to #pics \ Define picture file header
    #sector to #first  pic: ;       \ Start BITMAP-picture file

: >PICT     ( a u -- da )   \ Convert string a u to start address da of picture(s)
    'seek? if  100 um*  20 M+  then ;

: [FADR]    ( "name" -- )
    seek? 0= ?abort  100 um* 20 M+  postpone 2literal ; immediate

: [FSECT]   ( "name" -- )
    seek? 0= ?abort  postpone literal ; immediate

\ Recursive directory tree
v: inside definitions
value NEST \ Track indentation for subdirs
: INDENT    ( -- )          nest for  ch . emit  next ;

: (TREE)    ( #DID -- )
    begin
    dup 0 f@  dup -1 <> while                           \ New subdir found?
        b-b #hash = if                                  \ Yes, wanted subir?
            cr  #hash 0= if  ." Root:"  cr  then        \ Ok, is it in root DIR? then show
            >r  2 +to nest  indent                      \ Indent subdir(s)
            dup 2 + 0 fc@+ ftype space  r>              \ Print DIR-name
            #hash >r  to #hash  10 + dup >r  recurse    \ Check next DIR entry with new hash value
            drop  r>  r> to #hash   -2 +to nest         \ Restore DID, hash & unindent display
        else
            drop  10 +                                  \ To next DIR entry
        then
    repeat
    drop ;                                              \ Ready, clear stack

v: extra definitions
: TREE      ( -- )          0 to nest  0 to #hash  #DID (tree) drop ;

v: fresh definitions
shield STORE\  freeze
mount

\ End
