(*      W25Q16 16 Mbit, a 2048 kByte SPI Flash chip
        So there will be room for about 200 noForth binaries...
        The max. size of flash used with this code is 128 Mbit,
        that is 16384 kByte, e.g. a W25Q128

Instructie - Omschrijving       - Write data bytes      - Read data bytes

    01      Write status reg.   - lb hb
    02      Program page        - a2 a1 a0 d00 to dff   -
    03      Read data           - a2 a1 a0              - db
    04      Write disable       - Noop
    05      Read status reg. 1  - s0
    06      Write enable        - Noop
    20      4K sector erase     - a2 a1 a0
    35      Read status reg. 2  - s1
    50      Write enable status - Noop
    60      Chip erase          - Noop
    90      Manufacturer ID     - xx xx 00              - mf id
    AB      Power on            - xx xx xx              - id
    B9      Power down          - Noop
    FF      Reset cont. read    - FF

P2.3  - CS                      \ SPI enable low      x1=Select
P2.4  - CLOCKPULSE              \ Clock               x1=Clock
P2.6  - DATA-IN                 \ Data bitstream in   x0=Miso
P2.5  - DATA-OUT                \ Data bitstream out  x1=Mosi
PJ.1  - Led
PJ.0  - Switch S2

*)

\ : M+        ( d1 n -- d2 )              s>d d+ ;
code M+     ( d1 n -- d2 )      \ Mixed plus
    470A ,          \ tos sun mov
    4308 ,          \ #0 day mov
    930A ,          \ #0 sun cmp
    3401 , 4338 ,   \ >? if,  #-1 day mov  then,
    4437 ,          \ sp )+ tos mov
    5A84 , 0 ,      \ sun sp ) add
    6807 , next     \ day tos addc
end-code

v: inside also definitions
: SPI-OFF       ( -- )      70 20D *bic ; \ P2SEL1  3-wire SPI
: SPI-ON        ( -- )      70 20D *bis ; \ P2SEL1  3-wire SPI

: SPI-SETUP     ( -- )
    01 5E0 **bis    \ UCA1CTLW Reset eUSCI
    03 322 c!       \ PJOUT     Red led on, Pullup on PJ.1
    02 324 c!       \ PJDIR     PJ.1 is output
    01 326 *bis     \ PJREN     PJ.0  Resistor on
    70 20B *bic     \ P2SEL0    P2.4 P2.5 P2.6 is SPI
    spi-on          \           3-wire SPI
    0C 205 c!       \ P2DIR     P2.2 & P2.3, P2.4 for SPI
    C0 207 *bis     \ P2REN     P2.7 resistor on
    C0 203 *bis     \ P2OUT     P2.7 pullup
    A981 5E0 !      \ UCA1CTLW  Clk=low, MSB first, Master, Synch, 3-wire, SMClk
    01 5E6 !        \ UCA1BRW   Clock is 16Mhz/1 = 16 MHz
    01 5E0 **bic    \ UCA1CTLW  Free eUSCI
    08 203 *bis     \ P2OUT  cs = 1
    02 322 *bic ;   \ PJOUT  PJ.1  Red led off

code SPI-I/O    ( b1 -- b2 )            \ Read and write at SPI-bus
    B3E2 , 5FC ,        \ begin,  #2 5FC & bit  \ UCA1IFG
    2BFD ,              \ cs? until,
    47C2 , 5EE ,        \ tos 5EE & mov \ UCA1TXBUF
    B3D2 , 5FC ,        \ begin,  #1 5FC & bit  \ UCA1IFG
    2BFD ,              \ cs? until,
    4257 , 5EC ,        \ 5EC & tos mov \ UCA1RXBUF
    next
end-code

: SPI-OUT       ( b -- )    spi-i/o drop ;  \ Write x to SPI-bus
: SPI-IN        ( -- b )    0 spi-i/o ;     \ Read b from SPI-bus
: {SPI)         ( -- )      08 203 *bic ;   \ P2OUT  SPI on CS=low
: SPI}          ( -- )      08 203 *bis ;   \ P2OUT  SPI off CS=high


\ SPI Flash start here

v: inside definitions
\ Pointers & memory allocation
200000.  100 um/mod nip  constant #FLASH   \ Flash end for 16 Mbit Flash in sectors
200 constant #BUFFERS   \ Space for two sector buffers

value #SECTOR   \ Flash sector counter
value BUF       \ Buffer in use  0/1
value #N        \ Byte index in sector buffer
create 'SECTOR  #buffers allot \ Reserve sector buffers

: {SPI        ( b -- )  {spi)  spi-out ;
: WRITE-ON    ( -- )    6 {spi spi} ;
: STATUS1     ( -- b )  5 {spi  spi-in  spi} ;
: BUSY        ( -- )    begin  status1 1 and 0= until ;
: CHIP-ERASE  ( -- )    0 to #sector  write-on  60 {spi spi}  busy ;
: POWER-DOWN  ( -- )    B9 {spi spi}  spi-off ;
: POWER-UP    ( -- )    spi-on  AB {spi spi} ;

\ Read bytes, words & strings from flash
: ADDR-FLASH  ( da -- )        spi-out  b-b spi-out spi-out ;
: FC@         ( da -- b )      3 {spi  addr-flash  spi-in  spi} ;
: FC@+        ( da1 -- da2 b ) 2dup 2>r  1 M+  2r> fc@ ;
: FTYPE       ( da u --  )     for  fc@+ pchar emit  next  2drop ;
: F@          ( da -- x )      03 {spi addr-flash spi-in spi-in spi}  b+b ;
: F@+         ( da -- da x )   2dup 2>r  2 M+  2r> f@ ;

\ Read & write 256 byte sectors from and to flash & patch buffer
: ADDR-SECTOR   ( sa -- )       0 swap b-b  spi-out spi-out spi-out ;
: BUFFER        ( +b -- a )     1 and  100 *  'sector + ;

: ADJ-SECTOR    ( -- )      \ Adjust #SECTOR to 4 kByte sector
    #sector 10 mod  ?dup if  10 swap - +to #sector  then ;

: READ-SECTOR   ( sa +b -- )    \ Fill buffer +b from sa, the address of a 256 byte sector
    >r  03 {spi  addr-sector  r> buffer
    100 for  spi-in  over c!  1+  next  spi} drop ;

: WRITE-SECTOR  ( sa +b -- )    \ Write buffer +b to sa address of a 256 byte sector
    >r  write-on  02 {spi  addr-sector  r> buffer
    100 for  count spi-out  next  drop  spi}  busy ;

\ Erase goes in 4 kByte sectors
: ERASE-SECTOR  ( sa -- )       \ sa is a 256 byte sector address in a 4 kByte sector!
    write-on  20 {spi  addr-sector  spi}  busy ;

v: extra definitions
\ Some tools
\ : ID.           ( -- )          \ Read manufacturer & device id
\    90 {spi  0 addr-sector  spi-in spi-in  spi}  . . ;

: .DFREE        ( -- )
    base @ >r  decimal  #flash #sector -  4 / . ." kBytes "  r> base ! ;

\ Tools for SPI Flash memory
: FDUMP         ( da u -- )
    0 ?do
        cr  2dup d.str 5 rtype  ." : "      \ Print address
        10 0 do                             \ Dump 16 bytes
            2dup i m+ fc@ 2 .r space        \ Print 16 bytes in hex.
        loop
        ch | emit  2dup 10 ftype  ." | "    \ Print 16 bytes in visible ASCII
        10 m+   stop? if  leave  then       \ Adjust address & test for key
    10 +loop  2drop ;

v: fresh  
shield FLASH\  freeze

\ End
