forth: ssd1306setup64
(* E61a - For noForth FR5xxx C&V, I2C driver for SSD1306 0.91 inch 128x32
   pixels oled screen using USCI I2C routines. Separated files with a small
   and big character set.

                     MSP430FR5949
                  -----------------
              /|\|              XIN|-
               | |                 |
               --|RST          XOUT|-
                 |                 |
                 |             P1.6|-> SDA (UCB0SDA)
                 |                 |
                 |             P1.7|-> SCL (UCB0SCL)

 Addresses  - Labels       - Bit patterns
 1CC   15C  - WDTCL        - Off already
 20A   20A  - P1SEL        - 00C
 20C   20C  - P1SEL2       - 00C
 540   640  - UCB0CTLW0    - FF1
 542   642  - UCB0CTLW1    - 000
 546   646  - UCB0BRW      - 050
 548   648  - UCB0STATW    - sss USCI status
 54C   64C  - UCB0RXBUF    - rrr RX Data
 54E   64E  - UCB0TXBUF    - ttt TX Data
 554   654  - UCB0I2C0A    - ooo NC
 560   660  - UCB0I2CSA    - 042 /2
 56A   66A  - UCB0CIE      - 000 USCI interrupt enable
 56C   66C  - UCBOIFG      - 002 = TX ready, 001 = RX ready

Label    P1  P2  P3  Function
------------------------------------------------
PxIN     200 201 220 Input
PxOUT    202 203 222 Output
PxDIR    204 205 224 Direction
PxREN    206 207 226 Resistor enable
PxSEL0   20A 20B 22A Select 0
PxSEL1   20C 20D 22C Select 1
P1IV     20E 21E 22E Interrupt vector word
P1SELC   210 211 230 Complement selection
PxIES    218 219 238 Interrupt edge select
PxIE     21A 21B 23A Interrupt on
PxIFG    21C 21D 23C Interrupt flag

FR2433  - UCB0 P1.2 = SDA, P1.3 = SCL
FR2355  - UCB0 P1.2 = SDA, P1.3 = SCL, UCB1 P4.6 = SDA, P4.7 = SCL
FR5949  - UCB0 P1.6 = SDA, P1.7 = SCL
FR5969  - UCB0 P1.6 = SDA, P1.7 = SCL
FR5994  - UCB0 P1.6 = SDA, P1.7 = SCL, UCB1 P5.0 = SDA, P5.1 = SCL
        - UCB2 P7.0 = SDA, P7.1 = SCL, UCB3 P6.4 = SDA, P6.5 = SCL

*)

hex v: inside also definitions
40 constant #Y              \ Display size on Y-axis

\ code ZERO?  ( u1 x -- u2 x f | x t )  \ LOOP replacement counter
\    #1 sp ) sub  0=? if,
\        tos sp ) mov  #-1 tos mov  next
\    then,
\    tos sp -) mov  #0 tos mov  next
\ end-code
code ZERO?      ( u1 x -- u2 x f | x t ) \ LOOP replacement counter
    8394 , 0 ,  2005 ,  4784 , 0 ,
    4337 ,  next  8324 ,  4784 ,  0 ,  4307 ,  next
end-code

\ Graphic OLED screen 128x64 pixels
value INV?                \ Inverted display?
code INV    ( b1 -- b2 )  \ Invert b1 to leave b2
    adr inv? & tos .b bix  next
end-code
\ code INV    ( b1 -- b2 )  \ Invert b1 to leave b2
\    E257 ,  adr inv? ,  next
\ end-code

v: extra definitions
: WHITE     ( -- )          false to inv? ;         \ White on black display
: BLACK     ( -- )          true  to inv? ;         \ Black in white display
: {OL       ( b -- )        78 {i2write  i2out1 ;   \ Start an oled command: b=00 or old data: b=40
                                                    \ Single byte command: b=80, single byte data: b=C0
: OL}       ( b -- )        i2out  i2stop} ;        \ End an oled stream
: >ROW      ( b -- )        inv  C0 {ol  ol} ;      \ Single pixel row
: >BRIGHT   ( b -- )        00 {ol 81 i2out ol} ;   \ b = 0 to 255 (max. brightness)
: ON/OFF    ( flag -- )     1 and AE or 80 {ol ol} ; \ Display on/off
: INVERSE   ( flag -- )     1 and A6 or 80 {ol ol} ; \ Display black or white

\ : {CMD      ( b -- )        80 {ol  i2out ;
\ : CMD       ( b -- )        80 {ol  ol} ;           \ Single byte oled command
\ : 2CMD      ( b1 b0 -- )    00 {ol  i2out  ol} ;    \ Dual byte oled command

value X  value Y
\ value WAIT          \ Picture hold time
v: inside definitions
code ONSCR?     ( -- fl )   \ X and Y on existing screen location, leave true
    8324 ,  4784 ,  0 ,     \ tos sp -) mov
    90B2 ,  80 ,  adr x ,   \ 80 # adr x & cmp
    2C06 ,  92B2 ,  adr y , \ cc? if,   #8 adr y & cmp
    2C03 ,  4337 ,  next    \   cc? if,   #-1 tos mov  next  then,
    4307 ,  next            \ then, #0 tos mov  next
end-code

\ code ?ROWS  ( n -- rows )
\    #0 tos cmp >? if,       \ ABS, make positive
\       #-1 tos bix  #1 tos add
\    then,
\    begin,
\    #0 tos cmp  0<>? while, \ Rows not zero?
\        7F # day mov
\        adr x & day cmp  u>? while, \ X pointer off screen?
\            #1 tos sub  #1 adr x & add \ Correct X & rows
\    repeat,
\        then,  next
\ end-code
: ?ROWS     ( n -- rows ) \ Leave nr. of rows left on screen
    abs  begin  dup onscr? 0= and  while 1-  incr x  repeat ;

v: extra definitions
: XY)         ( -- )
    onscr? if               \ X & Y are valid?
        00 {ol  y B0 or i2out \ Yes, set page
        x  dup 0F and i2out \ & set column
        4 rshift 10 or ol}
    then ;

: XY           ( x y )      to y  to x  xy) ;

: SETUP-DISPLAY ( -- )
    setup-i2c               \ Init. 400kHz USCI I2C
    false on/off            \ Display off
    00 {ol                  \ Start oled-command stream
    0D5 i2out  080 i2out    \ Set oscillator clock
    0A8 i2out  03F i2out    \ Set multiplexer ratio = 1F
    0D3 i2out  000 i2out    \ Display offset = 0
    040 i2out               \ Display starts at line 0
    08D i2out  014 i2out    \ Charge pump on
    0A1 i2out               \ Mirror X-axis = A0
    0C8 i2out               \ Mirror Y-axis = C0
    0DA i2out  012 i2out    \ Alternate Com pin map = 02
    0D9 i2out  022 i2out    \ Set precharge cycles to high cap.
    0DB i2out  030 i2out    \ VCOMH voltage to max. = 40
    0A4 i2out               \ Enable rendering from GDRAM
    020 i2out  000 ol}      \ Horizontal display mode, end stream
    C0 >bright              \ Set brightness to 75%
    false inverse  white    \ Oled in normal mode
    true on/off ;           \ Display on

: &FILL         ( +n b -- ) \ Pattern 'b' to +n columns
    40 {ol              \ Start oled-data stream
    begin               \ Whole screen buffer
        dup i2out       \ Output pattern
    zero? until  drop  i2stop} ; \ End stream

: &EOL          ( u -- )        40 {ol  for  0 i2out  next  i2stop} ;

: .ROW      ( rows a -- )
    40 {ol  begin  count i2out  zero? until  drop  i2stop} ;

: .BITROW   ( a +n -- )     \ Output half of +Nx14 (big) character
    swap  40 {ol  0 i2out
    begin  count i2out  1+  \ Output every second bit row
    zero? until  drop  0 ol} ;

value O-EMIT                \ OLED emit vector
v: extra definitions
: &ERASE        ( -- )          400 0 &fill ;       \ Erase screen
: &HOME         ( -- )          0 0 xy ;            \ To upper left corner
: &PAGE         ( -- )          &erase &home ;
: &CR           ( -- )          0 y 2 + xy ;        \ For 16-bit chars
: &EMIT         ( c -- )        o-emit execute ;
: &SPACE        ( -- )          bl &emit ;
: &SPACES       ( u -- )        for  &space  next ;
: &TYPE         ( a u -- )      bounds ?do  i c@ &emit  loop ;
: &U.           ( n -- )        0 d.str &type &space ;
: WAIT          ( u -- )        for  10 ms  key? if rdrop exit then  next ;

(*
: &XTYPE        ( xa +n -- )    0 ?do  xc@+ &emit  loop  2drop ; \ String from XRAM
: X&TYPE        ( a u -- )      bounds 0 ?do  10000. i m+ xc@ &emit  loop ;
: X"(           ( -- )          10000. inl# m+  xc@+ x&type ;
: "(            ( -- )          inl# count &type ;
: &"            ( -- )          \ ." voor OLED
    flyer  postpone x"(  xhere drop ,
    ch " parse dup xc,  bounds
    ?do i c@ xc, loop  xalign ;  immediate
*)

: &"            ( -- )          \ ." voor OLED
    flyer  postpone s"  postpone &type ;  immediate

: XY"       ( x y ccc -- )      \ XY" voor OLED
    flyer  postpone xy  postpone &" ; immediate

code C>N        ( c -- +n )     \ Convert char to bitmap index number
    9037 ,  80 ,  2802 ,        \ 80 # tos cmp  cs? if,
    4037 ,  80 ,  8037 ,  20 ,  \ 80 # tos mov  then,  20 # tos sub
    next
end-code

v: fresh definitions 
shield SSD1306\  freeze
setup-display       \ OLED active
&page               \ With an empty display

\ End ;;;
